/**
 * @file
 * This is part of the first tile tutorial.  Information was gather
 * from several locations to get this to work.  At the minimum, the
 * following sites were consulted to get this far:
 *
 * - http://www.drunkencoders.com/
 * - http://www.webbesen.dk/gba/
 *
 * This code has been tested in both dualis realease 2 as well as on
 * an actual Nintendo DS via a PassMe device.
 *
 * PhoenixRising
 **/

typedef unsigned short int u16;

#include "KeyPad.hpp"
#include "alpha8.h"
#include <NDS/NDS.h>
#include <NDS/ARM9/BIOS.h>

/**
 * @brief
 * The class that holds all of the data related to the application.
 * The main simply instances one of these classes and then executes it
 * via the "go" method.
 **/
class Tutorial01Tile {
public:
  /**
   * @brief
   * Mechanism to get ahold of the application.  This ensures that
   * both main and the interrupt handler can gain access to the same
   * application object without having to resort to global variables
   * and other evilness.
   *
   * @retval A reference to the application object
   **/
  static Tutorial01Tile& theApp(void);

  /**
   * @brief
   * Gets everything on the screen, and then goes to sleep, letting
   * the interrupt handler update things appropriately.
   **/
  void go(void);

  /**
   * @brief
   * Updates the display.
   **/
  void updateDisplay(void);

  void paintKeyboard(bool shifted);

private:
  /**
   * @brief
   * Constructor prepares the application to run.  To access an
   * object, use theApp() to gain a reference to the application.
   *
   * It does the following:
   * - Initializes the internal class variables.
   * - Initializes the palette to black.  Not required, but makes
   *   viewing the palette in dualis easier.
   * - Configures the various registers on the DS to initialize both
   *   displays in MODE_0_2D.
   * - Finally, interrupts are enabled with the default 
   **/
  Tutorial01Tile();

  /**
   * @brief
   * Copy constructor is disabled.
   **/
  Tutorial01Tile(const Tutorial01Tile&);

  /**
   * @brief
   * Operator= is disabled.
   **/
  Tutorial01Tile& operator=(const Tutorial01Tile&);

  /**
   * @brief
   * Destructor never gets called because the application never ends.
   **/
  virtual ~Tutorial01Tile();  

  /**
   * @brief
   * Determines if the keyboard caps-lock key is on.
   **/
  bool shifted;

  /**
   * @brief
   * Current X location on the screen.
   **/
  int x;

  /**
   * @brief
   * Current Y location on the screen.
   **/
  int y;

  /**
   * @brief
   * Notes if the cursor is currently on the screen.
   **/
  bool cursor_displayed;

  /**
   * @brief
   * Flash delay.  How many updates between turning the cursor on/off.
   **/
  int flash_delay;

  /**
   * @brief
   * Tile map for the top screen.
   **/
  u16* top_map;

  /**
   * @brief
   * Tile map for the bottom screen.
   **/
  u16* bottom_map;

  /**
   * @brief
   * The key to use for a cursor.
   **/
  char cursor;

  /**
   * @brief
   * data for rows of keyboard without shift.
   **/
  char* rows[4];

  /**
   * @biref
   * data for rows of keyboard with shift.
   **/
  char* shifted_rows[4];
};

/**
 * @brief
 * 
 **/
static void InterruptHandler(void) {
  if (IF & IRQ_VBLANK) {
    Tutorial01Tile& app(Tutorial01Tile::theApp());

    app.updateDisplay();

    IF = IRQ_VBLANK;
  }
}

void Tutorial01Tile::paintKeyboard(bool shifted)
{
  char** d = (shifted) ? shifted_rows : rows;

  int row, column, map_pos;;
  for(row = 0; row < 4; ++row) {
    for(column = 0; d[row][column] != '\0'; ++column) {
      map_pos = ((row * 4 + 5) * 32) + (column * 2);
      bottom_map[map_pos] = d[row][column] - ' ';
    }
  }
}

/* VRAM Bank Control */
#define VRAM_ALL_ENABLE    (1 << 7)
 
//do not combine these defines (other than offset for main display)
#define VRAM_ALL_LCD       0
#define VRAM_ABCD_BG       1
#define VRAM_ABCD_OBJ      2
#define VRAM_ABCD_TEXTURE  3
 
Tutorial01Tile::Tutorial01Tile()
  : shifted(false), x(0), y(0), cursor_displayed(true), flash_delay(15),
    top_map(0), bottom_map(0), cursor('^')
{
  unsigned int i;

  for(i = 0; i < 0x200; ++i) {
    BG_PALETTE[i]=0;
    SPRITE_PALETTE[i] = 0;
    BG_PALETTE_SUB[i]=0;
    SPRITE_PALETTE_SUB[i] = 0;
  }

  // Turn on the screens and 2D cores and switch to mode 0

  POWER_CR = POWER_ALL_2D | POWER_MATRIX;
 
  vramSetMainBanks(VRAM_A_MAIN_BG, VRAM_B_MAIN_BG,
		   VRAM_C_SUB_BG, VRAM_D_SUB_SPRITE);

  DISPLAY_CR = MODE_0_2D | DISPLAY_BG0_ACTIVE | DISPLAY_BG1_ACTIVE | OBJ_MAP_2D;
  SUB_DISPLAY_CR = MODE_0_2D | DISPLAY_BG0_ACTIVE | DISPLAY_BG1_ACTIVE | OBJ_MAP_2D;

  // Load our tile information
  for(u16 i = 0; i < sizeof(alpha8Palette) / sizeof(u16); ++i) {
    BG_PALETTE[i] = BG_PALETTE_SUB[i] = alpha8Palette[i];
  }

  SUB_BG0_CR = BG_256_COLOR | (0 << CHAR_SHIFT) | (20<< SCREEN_SHIFT);
  BG0_CR = BG_256_COLOR | (0 << CHAR_SHIFT) | (20<< SCREEN_SHIFT);

  u16* sub_tile = (u16*)CHAR_BASE_BLOCK_SUB(0);
  top_map = (u16*)SCREEN_BASE_BLOCK_SUB(20);
  u16* tile = (u16*)CHAR_BASE_BLOCK(0);
  bottom_map = (u16*)SCREEN_BASE_BLOCK(20);

  BG_PALETTE[0] = BG_PALETTE_SUB[0] = RGB15(0, 0, 0);
  BG_PALETTE[1] = BG_PALETTE_SUB[1] = RGB15(0, 31, 0);

  for(i = 0; i < sizeof(alpha8Data) / sizeof(u16); ++i) {
    tile[i] = sub_tile[i] = alpha8Data[i];
  }

  rows[0] = "`1234567890-=";
  rows[1] = " qwertyuiop[]\\";
  rows[2] = " asdfghjkl;'";
  rows[3] = " zxcvbnm,./";
  shifted_rows[0] = "~!@#$%^&*()_+";
  shifted_rows[1] = " QWERTYUIOP{}|";
  shifted_rows[2] = " ASDFGHJKL:\"";
  shifted_rows[3] = " ZXCVBNM<>?";

  paintKeyboard(shifted);
  top_map[x + y * 32] = cursor - ' ';

  // Enable the V-blank interrupt
  IME = 0;
  IRQ_HANDLER = &InterruptHandler;
  IE = IRQ_VBLANK;
  IF = ~0;
  DISP_SR = DISP_VBLANK_IRQ;
  IME = 1;
}

Tutorial01Tile::~Tutorial01Tile()
{}

Tutorial01Tile& Tutorial01Tile::theApp(void)
{
  static Tutorial01Tile app;
  return app;
}

void Tutorial01Tile::go(void)
{
  while (1) {
    swiWaitForVBlank();
  }
}

void writeString(const char* str, int x, int y, u16* map_base)
{
  map_base += x + y * 32;
  while(*str != '\0') {
    *map_base = *str - ' ';
    ++map_base;
    ++str;
  }
}

void writeNumberHelper(u16 num, int places, char* dest)
{
  if (places) {
    if (num) {
      writeNumberHelper(num / 10, places - 1, dest + 1);
    }
    *dest = (num % 10) + '0';
    if (!num) {
      --places;
      ++dest;
      while(places) {
	*dest = ' ';
	--places;
	++dest;
      }
    }
  }
}

void writeNumber(u16 num, int x, int y, u16* map_base)
{
  char tmp[6] = {'\0'};
  writeNumberHelper(num, sizeof(tmp) - 1, tmp);
  writeString(tmp, x, y, map_base);
}

void writeNumberHex(u16 num, int x, int y, u16* map_base)
{
  static char char_index[16] = {
    '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', 'A', 'B', 'C', 'D', 'E', 'F',
  };
  char tmp[7] = {'\0'};
  tmp[6] = '\0';
  tmp[5] = char_index[num & 0x000F]; num = num >> 4;
  tmp[4] = char_index[num & 0x000F]; num = num >> 4;
  tmp[3] = char_index[num & 0x000F]; num = num >> 4;
  tmp[2] = char_index[num & 0x000F];
  tmp[1] = 'x';
  tmp[0] = '0';

  writeString(tmp, x, y, map_base);
}

void writeNumberBinary(u16 num, int x, int y, u16* map_base)
{
  int i;
  char tmp[22] = {'\0'};
  char* dest = &tmp[sizeof(tmp) - 2];
  for(i = 1; i <= 16; ++i) {
    *dest = (num & 0x0001) + '0';
    --dest;
    if ((i % 4) == 0) {	// Space between nibbles.
      *dest = ' ';
      --dest;
    }
    num = num >> 1;
  }
  tmp[0] = '0';
  tmp[1] = 'b';
  writeString(tmp, x, y, map_base);
}

void debugTouchScreen(u16* map)
{
  int rawX = IPC->touchX;
  int rawY = IPC->touchY;
//   int xLoc = (rawX - 0x0113) / 14;
//   int yLoc = (rawY - 0x00E0) / 19;
//   int keyX = xLoc / 8;
//   int keyY = yLoc / 8;

  writeString("raw", 2, 21, map);
  //writeString("Loc", 12,0, map);
  //writeString("key", 20,0, map);
  writeString("X:", 0, 22, map);
  writeString("Y:", 0,  23, map);
  writeNumberBinary(rawX, 2,  22, map);
  writeNumberHex(rawX, 25,  22, map);
  //writeNumber(xLoc, 12, 22, map);
  //writeNumber(keyX, 20, 22, map);
  writeNumberBinary(rawY, 2,  23, map);
  writeNumberHex(rawY, 25,  23, map);
  //writeNumber(yLoc, 12, 23, map);
  //writeNumber(keyY, 20, 23, map);
}

void Tutorial01Tile::updateDisplay(void) {
  static KeyPad kp(20, 5);
  static int counter = 0;
  static bool touched = false;
  char display_char = '_';
  uint16 keypress;
  int dx = 0;
  int dy = 0;

  keypress = kp.Update();
  ++counter;

  if (keypress & KEY_UP) {
    dy = -1;
  }
  if (keypress & KEY_DOWN) {
    dy = 1;
  }
  if (keypress & KEY_LEFT) {
    dx = -1;
  }
  if (keypress & KEY_RIGHT) {
    dx = 1;
  }
  if (keypress & KEY_L) {
    shifted = !shifted;
    paintKeyboard(shifted);
  }

  if (((counter == flash_delay) || (dx != 0) || (dy != 0)) &&
      (cursor_displayed)) {
    counter = 0;
    display_char = ' ';
    top_map[x + y * 32] = display_char - ' ';
    cursor_displayed = !cursor_displayed;
  }

  debugTouchScreen(top_map);

  bool pen_down = (~(IPC->buttons) & (1 << 6)) ? true : false;
  if (pen_down) {
    if (!touched) {
      int rawX = IPC->touchX;
      int rawY = IPC->touchY;
      int xLoc = (rawX - 0x0113) / 14;
      int yLoc = (rawY - 0x00E0) / 19;
      int keyX = xLoc / 8;
      int keyY = yLoc / 8;

      if (bottom_map[keyX + keyY * 32] != 0) {
	dx += 1;
	top_map[x + y * 32] = bottom_map[keyX + keyY * 32];
      } else {
	keyX = (xLoc - 2) / 8;
	keyY = (yLoc - 2) / 8;
	if (bottom_map[keyX + keyY * 32] != 0) {
	  dx += 1;
	  top_map[x + y * 32] = bottom_map[keyX + keyY * 32];
	} else {
	  keyX = (xLoc + 2) / 8;
	  keyY = (yLoc + 2) / 8;
	  if (bottom_map[keyX + keyY * 32] != 0) {
	    dx += 1;
	    top_map[x + y * 32] = bottom_map[keyX + keyY * 32];
	  }
	}
      }
    }
  }

  touched = pen_down;

  x += dx;
  y += dy;

  if (x > 31) {
    x = 0;
  }

  if (x < 0) {
    x = 31;
  }

  if (y < 0) {
    y = 23;
  }

  if (y > 23) {
    y = 0;
  }

  if (((counter == flash_delay) || (dx != 0) || (dy != 0)) &&
      (!cursor_displayed)) {
    counter = 0;
    display_char = cursor;
    top_map[x + y * 32] = display_char - ' ';
    cursor_displayed = !cursor_displayed;
  }
}

int main(int argc, char** argv) {
  Tutorial01Tile& app(Tutorial01Tile::theApp());

  app.go();

  return 0;
}


